/***************************************************************************
                          game_client.cpp  -  description
                             -------------------
    begin                : Thu Feb 3 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "game_client.h"
#include "all.h"

//-----------------------------------------------------------------------------
Game_client::Game_client()
{
	// This must happen before setgeom()!
	nplayers=0;
	for(int i=0; i < MAX_PLAYERS; ++i)
	  {
	    players[i].player_name[0]='\0';
	    players[i].viewport=NULL;
	    players[i].controls.owner=i;
	    players[i].owner=-1;
	    players[i].joystick=i;

	    players[i].elevate_up = false;
	    players[i].elevate_down = false;
	  }

	last_tick_time = 0;
	last_tick_duration = 1;
	loader = NULL;
	dropshadow = true;
	smooth = false; // DAMMIT!
	bars_at_top = false;
	smallfont = NULL;
	setgeom(0,0,320,240,2);
}

//-----------------------------------------------------------------------------
Game_client::~Game_client()
{
	if (tcp_client.send(NETSTATE::QUIT) < 0)
		logmsg(lt_error,"unable to tell the server we want to quit.");

	for(int i=0; i<nplayers; ++i)
	  {
	    delete players[i].viewport;
	  }
}

//-----------------------------------------------------------------------------
//call to get things going:
int Game_client::init(Loader *loader_, Soundsystem *soundsystem_,
		      Configuration *configuration_)
{
	loader = loader_;
	soundsystem = soundsystem_;
	config_attach(configuration_);
	config_changed();
	return 0;
}

//-----------------------------------------------------------------------------
// Just checks who's using what joysticks.
void Game_client::config_changed()
{
	players[0].joystick=configuration->get_option_int("Player1_joystick");
	players[1].joystick=configuration->get_option_int("Player2_joystick");
	players[2].joystick=configuration->get_option_int("Player3_joystick");
	players[3].joystick=configuration->get_option_int("Player4_joystick");

	for(int i=0; i<MAX_PLAYERS; ++i)
	  logmsg(lt_debug, "Player %i is using joystick %i",
		 i, players[i].joystick);
}

//-----------------------------------------------------------------------------
void Game_client::setup_viewports()
{
  if(nplayers==1)
    {
      players[0].xpos=0;
      players[0].ypos=0;

      delete players[0].viewport;
      players[0].width=width;
      players[0].height=height;
      players[0].viewport=new Viewport(width/view_quality,
				       height/view_quality,
				       players[0].owner);
    }
  else if(nplayers==2)
    {
      int w=width/2-2;

      players[0].xpos=0;
      players[0].ypos=0;
      players[0].width=w;
      players[0].height=height;

      delete players[0].viewport;
      players[0].viewport=new Viewport(w/view_quality,
				       height/view_quality,
				       players[0].owner);

      players[1].xpos=w+4;
      players[1].ypos=0;
      players[1].width=w;
      players[1].height=height;

      delete players[1].viewport;
      players[1].viewport=new Viewport(w/view_quality,
				       height/view_quality,
				       players[1].owner);
    }
  else if(nplayers>=3)
    {
      int w=width/2-2,h=height/2-2;

      players[0].xpos=0;
      players[1].xpos=w+4;
      players[2].xpos=0;

      if(nplayers==4)
				players[3].xpos=w+4;

      players[0].ypos=0;
      players[1].ypos=0;
      players[2].ypos=h+4;

      if(nplayers==4)
	players[3].ypos=h+4;

      for(int i=0; i<nplayers; ++i)
	{
	  players[i].width=w;
	  players[i].height=h;

	  delete players[i].viewport;
	  players[i].viewport=new Viewport(w/view_quality,h/view_quality, players[i].owner);
	}
    }

  for(int i=0; i<nplayers; ++i)
    {
      players[i].health_bar.setgeom(5,players[i].height-(7+5),
				    100*players[i].width/640,7);
      players[i].health_bar.bar_color = PIXELA(128,0,255,8);
      players[i].health_bar.render();
      players[i].viewport->setmapsize(map_xsize, map_ysize);
    }
}

//-----------------------------------------------------------------------------
int Game_client::setgeom(int xpos_, int ypos_, int view_xsize, int view_ysize, int view_quality_)
{
	xloc = xpos_;
	yloc = ypos_;
	width=view_xsize;
	height=view_ysize;
	view_quality = view_quality_;

	setup_viewports();
	return 0;
}

// join_game only requests the map from the server now; the player
// addition is done elsewhere.

//-----------------------------------------------------------------------------
int Game_client::join_game(const char *server, int server_port,
			   int nplayers_,
			   const char *name, int color)
{
        nplayers=nplayers_;
	// FIXME: use different names/colors for each player.
	for(int i=0; i<nplayers; ++i)
	  {
	    strcpy(players[i].player_name,
		   name);
	    players[i].player_color=color;
	    players[i].viewport=NULL;
	  }

	{
		unsigned int firsttry = time_ms();
		int error = -1;

		logmsg(lt_debug,"Trying to connect to server %s:%i",server,server_port);
		while (error < 0 && time_elapsed(firsttry,time_ms()) < 10000) {
			error = tcp_client.connect(server,server_port);
			if (error < 0) {
				sleep(1);
				logmsg(lt_debug,"Couln't connect, but trying again...");
			}
		}

		if (error < 0) {
			logmsg(lt_error,"Unable to get connected to the server, quitting. %i",error);
			return -1;
		} else
			logmsg(lt_message,"Yes!, we are connected to the server!");
	}

	{
		Net_server_signature server_signature;
		if (tcp_client.recv(&server_signature) < 0) {
			logmsg(lt_error,"Unable to get server signature");	
			return -2;
		}

    if (!server_signature.server_ok()) {
			logmsg(lt_fatal,"Server is not a NiL server, it said: %s",server_signature.signature);
			return -3;
		};

    if (!server_signature.version_ok()) {
			logmsg(lt_fatal,"Server is a NiL server, but it has the wrong version: %s you are using %s",server_signature.version,PROTOCOL_VERSION);
			return -4;
		};
	}

	Net_joinrequest joinrequest;

	joinrequest.nplayers=nplayers;

	for(int i=0; i<nplayers; ++i)
	  {
	    strcpy(joinrequest.players[i].name,
		   players[i].player_name);
	    joinrequest.players[i].color=players[i].player_color;
	  }

	if (tcp_client.send(&joinrequest) < 0) {
	  logmsg(lt_error,"tcp_client.send(&joindata,sizeof(joindata) failed.");
	  return -5;
	}
	logmsg(lt_debug,"join request sent");

	//get join reply
	Net_joinreply joinreply;
	if (tcp_client.recv(&joinreply) < 0) {
	  logmsg(lt_error,"tcp_client.recv(&joinreply,sizeof(joinreply)) failed.");
	  return -6;
	}
	logmsg(lt_debug,"got join reply from server.");

	for(int i=0; i<nplayers; ++i)
	  {
	    logmsg(lt_debug, "Player %i has id %i", i, joinreply.player_ids[i]);
	    players[i].owner=joinreply.player_ids[i];
	  }

	//create client world
	//	Clientworld clientworld(&loader,joinreply.objectid);
	clientworld.init(loader, soundsystem);
	logmsg(lt_debug,"client world created, about to initmap.");				

	{
		//get the compressed map:
		char *map_data = (char*)malloc(joinreply.compressed_size);
 	 	if (tcp_client.recv(map_data,joinreply.compressed_size) < 0) {
			logmsg(lt_error,"Couldn't read compressed map diff.");
			return -7;
		}
		logmsg(lt_debug,"got map.");				

		//have the clientworld unpack the packed gob:
		clientworld.createmap(joinreply,map_data);
		free(map_data);
		logmsg(lt_debug,"initmap done.");
  }

	map_xsize=joinreply.map_xsize;
	map_ysize=joinreply.map_ysize;

	//send handshake:
	int handshake = 42;
  if (tcp_client.send(&handshake,sizeof(handshake)) < 0) {
		logmsg(lt_error,"Couldn't send handshake.");
		return -8;
	}

	setup_viewports();

	logmsg(lt_debug,"ready to go.");
	return 0;
}


//-----------------------------------------------------------------------------
//void Game_client::get_join_status(char *text, int textsize, int &progress_, int &max_progress_); remember to have a semaphor

//-----------------------------------------------------------------------------
//poll once for each frame:
int Game_client::think()
{
	int netstate = -1;
	int this_update_id = -1;

	//Get the update id:
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::UPDATEID) {
		logmsg(lt_error,"protocol error 1: got netstate %i.", netstate);
		return -1;
	}
	if (tcp_client.recv(this_update_id) <0) {
		logmsg(lt_error,"unable to get the update_id from the server.");
	}

	//Get the world update itself
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::WORLDUPDATE) {
		logmsg(lt_error,"protocol error 2.");
		return -1;
	}
	if (tcp_client.recv(&clientworld.serializer_writer) < 0) {
		logmsg(lt_error,"unable to get the worldupdate from the server.");
		return -1;
	}

	//Get the done flag:
	tcp_client.recv(netstate);
	if (netstate != NETSTATE::DONE) {
		logmsg(lt_error,"protocol error 3.");
		return -1;
	}


	//Send ack for this update:
	if (tcp_client.send(NETSTATE::UPDATEID) < 0) {
		logmsg(lt_error,"unable to send state to the server.");
		return -2;  	
	}
	if (tcp_client.send(this_update_id) < 0) {
		logmsg(lt_error,"unable to send this_update_id to the server.");
		return -2;  	
	}

	for(int i=0; i<nplayers; ++i)
	  {
	    //send the current controls
	    if (tcp_client.send(NETSTATE::CONTROLS) < 0) {
	      logmsg(lt_error,"unable to send state to the server.");
	      return -2;  	
	    }
	    if (tcp_client.send(&players[i].controls) < 0) {
	      logmsg(lt_error,"unable to send controls to the server.");
	      return -2;
	    }
	  }

	//Tell the server that there is no more lumps for now
	if (tcp_client.send(NETSTATE::DONE) < 0) {
		logmsg(lt_error,"unable to send state to the server.");
		return -2;  	
	}


  //update some statistics:
	last_update_size = clientworld.serializer_writer.getlength();
	last_tick_duration = clientworld.gettime()-last_tick_time;
	last_tick_time = clientworld.gettime();
	if (last_tick_duration > 10) {
		last_tick_duration = 1;
		trend_tick_duration = 1;
	}
	trend_tick_duration += (last_tick_duration-trend_tick_duration)/20;


	clientworld.client_think();

	controls_tick();

	return 0;
}

//-----------------------------------------------------------------------------
//call to have this game client draw its view of the world
int Game_client::draw(Mutable_raw_surface *target)
{
//	Obj_player_avatar *avatar = clientworld.get_player_avatar(clientworld.getowner());
  for(int i=0; i<nplayers; ++i)
    {
	Obj_player_data *player_data	= (Obj_player_data *)clientworld.get_object(players[i].owner,ot_player_data);
	if (!player_data) {
  	logmsg(lt_warning,"Unable to find player_data");
		return -1;
	}

	if ( smallfont )
		if ( !(players [ 0 ].frag_info.is_font()) ) {
			for ( int i = 0; i < nplayers; ++i ) {
				players[i].frag_info.set_font ( smallfont );
				for ( int i2 = 0; i2 < 3; ++i2 ) {
					players[i].message[i2].set_font ( smallfont );
        }
			}
		}
	Obj_player_avatar *avatar 		= player_data->get_player_avatar();
	//figure out if we are to put the bars at the top or the bottom:
	if (avatar) {
		float x,y;
		avatar->getpos(x,y);
		int switch_margin = players[i].viewport->get_ysize()/2;

    if (bars_at_top) {

    	if (y < switch_margin)
				bars_at_top = false;

		} else {

    	if (y > clientworld.map_buffer.get_ysize()-switch_margin)
				bars_at_top = true;

		}

	}

	//Figure out the pixel position of the bars
	int bars_ypos = bars_at_top?5:players[i].viewport->get_ysize()-(7+5);				

  //Update the health bar:
	if (avatar) players[i].health_bar.setpos((avatar->gethealth()*1.0)/MAX_HEALTH);
	players[i].health_bar.ypos = bars_ypos;

	//Update the weapon bars:

	if (player_data) {
	  // bar_width is calculated as follows: each bar has 5 pixels of
	  // padding on each side, and the bars are packed into the right-hand
	  // 3/4 of the screen.
		float spacing=10.0*players[i].width/640;
		float bar_start = 120.0*players[i].width/640;
 		float bar_width = 25*players[i].width/640;
		for (int s=0;s<WEAPON_SLOTS;s++) {
		  // FIXME: use separate controls for each player.
      float this_bar_width=players[i].controls.get_weapon()==s?bar_width*2:bar_width;
      players[i].weapon_bar[s].setgeom((int)bar_start,bars_ypos,ROUND(this_bar_width),7);

			Obj_weapon *weapon = player_data->get_weapon(s);
			if (weapon) {

				//Figure out the color of the bar
				int32 newcolor = weapon->canfire()?PIXELA(0,255,0,6):PIXELA(255,0,0,6);
				if (players[i].weapon_bar[s].bar_color != newcolor) {
					players[i].weapon_bar[s].bar_color = newcolor;
					players[i].weapon_bar[s].render();
				}
				
				float readyness = weapon->getreadyness();
				players[i].weapon_bar[s].setpos(readyness);
			} else {
				players[i].weapon_bar[s].setpos(1);
				logmsg(lt_warning,"Unable to find the weapon for slot %i",s);
			}

			bar_start += this_bar_width+spacing;
		}
		// Now render texts
    Obj_scoreboard *sb = clientworld.getscoreboard ();
    int id = get_player_id(i);
    Obj_scoreboard::Score sc = sb -> getscore ( id );
		if ( players[i].frag_info.is_font() ) {
			char str [ 100 ];
      if ( id != -1 ) {
        sprintf ( str , "Frags:%d, Kills:%d, Deaths:%d" , sc.frags , sc.kills , sc.deaths );
        players[i].frag_info.setgeom ( 10 , 10 , str , 255 , 255 , 255 );
        players[i].frag_info.render( );
      }
		}
    for ( int i2 = 0; i2 < sb -> frags_inuse; ++i2 )
      if ( players[i].message[i2].is_font() ) {
        char str [ 100 ];
        if ( id != -1 ) {
          Obj_player_data *pk = (Obj_player_data *)clientworld.get_object(sb -> last_killers[i2],ot_player_data);
          Obj_player_data *pv = (Obj_player_data *)clientworld.get_object(sb -> last_victims[i2],ot_player_data);
          if (( pk ) && ( pv )) {
            if ( sb -> last_killers [ i2 ] == sb -> last_victims [ i2 ] )
              sprintf ( str , "%s commited suicide" , pk -> get_player_name () );
            else
              sprintf ( str , "%s massacred %s" , pk -> get_player_name () , pv -> get_player_name () );
            int color = (int)((sb -> keep_frag_time [ i2 ]) - clientworld.gettime ()) * 128 / KEEP_FRAG_TIME + 128;
            players[i].message[i2].setgeom ( 10 , 10 * (i2 + 2) , str , color , color , color );
            players[i].message[i2].render ();
          }
        }
      }
	}

	//figure out what to view
	int vx = 0;
	int vy = 0;
	clientworld.get_player_pos(players[i].viewport->getowner(),vx,vy);
	players[i].viewport->addsample(vx,vy,last_tick_duration);

	clientworld.draw(players[i].viewport);

	if (dropshadow)
	  players[i].viewport->drop_shadow();

	//draw the player specific UI bits:
	players[i].frag_info.draw(players[i].viewport);
  Obj_scoreboard *sb = clientworld.getscoreboard ();
  for ( int i2 = 0; i2 < sb -> frags_inuse; ++i2 )
    players[i].message[i2].draw(players[i].viewport);
	players[i].health_bar.draw(players[i].viewport);
	for (int s=0;s<WEAPON_SLOTS;s++)
	  players[i].weapon_bar[s].draw(players[i].viewport);

   if (smooth) {
     players[i].viewport->draw_copy_stretch(
        target,
        0, 0,
        players[i].viewport->get_xsize(), players[i].viewport->get_ysize(),
        players[i].xpos+xloc, players[i].ypos+yloc,
        players[i].width, players[i].height,
        Raw_surface::OPAQUE);
   } else {
     players[i].viewport->draw_copy_stretch(
        target,
        0,0,
        players[i].viewport->get_xsize(), players[i].viewport->get_ysize(),
        players[i].xpos+xloc, players[i].ypos+yloc,
        view_quality
        );
   }
	
    }

  return 0;
}

//-----------------------------------------------------------------------------
// FIXME: this belongs on the server, doesn't it?  (think hacked client)
void Game_client::controls_tick()
{
  for(int i=0; i<nplayers; ++i)
    {
      if (players[i].elevate_down)
	players[i].controls.set_elevation(players[i].controls.get_elevation()-0.1);
      if (players[i].elevate_up)
	players[i].controls.set_elevation(players[i].controls.get_elevation()+0.1);	
    }
  controls_ticked = true;
}

//-----------------------------------------------------------------------------
bool Game_client::handle_input(Nil_inputevent *inputevent, int player)
{
	if (!inputevent || !inputevent->sdl_inputevent) return false;

        Controls &controls=players[player].controls;

	bool pressed = inputevent->pressed;
	int function = inputevent->cmdfunction & CMDFUNCTION::FUNCTION_MASK;

	switch (function) {

  	case (CMDFUNCTION::UP) :{
			if (pressed) {
				players[player].elevate_up = true;
				controls_ticked = false;
			} else {
				if (!controls_ticked) controls_tick();
				players[player].elevate_up = false;
			}
			return true;
		}

        case (CMDFUNCTION::DOWN) :{
			if (pressed) {
				players[player].elevate_down = true;
				controls_ticked = false;
			} else {
				if (!controls_ticked) controls_tick();
				players[player].elevate_down = false;
			}
			return true;
		}

    case (CMDFUNCTION::LEFT) :{
			controls.clear_flag(Controls::cf_direction_east);
			controls.set_flag(Controls::cf_walk,pressed);
			return true;
		}

		case (CMDFUNCTION::RIGHT) :{
			controls.set_flag(Controls::cf_direction_east);
			controls.set_flag(Controls::cf_walk,pressed);
			return true;
		}

		case (CMDFUNCTION::JUMP) :{
			controls.set_flag(Controls::cf_jump,pressed);
			return true;
		}

		case (CMDFUNCTION::HOOK_FIRE) :{
     	controls.set_flag(Controls::cf_hook_fire,pressed);
			return true;
		}

		case (CMDFUNCTION::HOOK_RELEASE) :{
			controls.set_flag(Controls::cf_hook_release,pressed);
			return true;
		}

		case (CMDFUNCTION::HOOK_IN) :{
			controls.set_flag(Controls::cf_hook_in,pressed);
			return true;
		}

		case (CMDFUNCTION::HOOK_OUT) :{
			controls.set_flag(Controls::cf_hook_out,pressed);
			return true;
		}

		case (CMDFUNCTION::FIRE) :{
			controls.set_flag(Controls::cf_shoot,pressed);
			return true;
		}

		case (CMDFUNCTION::DIG) :{
			controls.set_flag(Controls::cf_dig,pressed);
			return true;
		}

		case (CMDFUNCTION::WEAPON1) :{
			if (pressed)
       	controls.set_weapon(0);
			return true;
		}

		case (CMDFUNCTION::WEAPON2) :{
			if (pressed)
       	controls.set_weapon(1);
			return true;
		}

		case (CMDFUNCTION::WEAPON3) :{
			if (pressed)
       	controls.set_weapon(2);
			return true;
		}

		case (CMDFUNCTION::WEAPON4) :{
			if (pressed)
       	controls.set_weapon(3);
			return true;
		}

		case (CMDFUNCTION::WEAPON5) :{
			if (pressed)
       	controls.set_weapon(4);
			return true;
		}

	        case (CMDFUNCTION::NEXTWEAPON) :{
		  if(pressed)
		    controls.set_weapon((controls.get_weapon()+1)%5);
		  return true;
		}

	        case (CMDFUNCTION::PREVWEAPON) :{
		  if(pressed)
		    { // shouldn't (-1)%5==4?
		      if(controls.get_weapon()<=0)
			controls.set_weapon(4);
		      else
			controls.set_weapon(controls.get_weapon()-1);
		    }
		  return true;
		}

    default: {
			if(inputevent->sdl_inputevent->type==SDL_KEYUP ||
			   inputevent->sdl_inputevent->type==SDL_KEYDOWN) {
			  SDLKey sdlkey    = ((SDL_KeyboardEvent *) inputevent->sdl_inputevent)->keysym.sym;
			  logmsg(lt_error,"Unhandled key: 0x%x (%s) -> 0x%x",sdlkey,SDL_GetKeyName(sdlkey),function);
			}
			return false;
	     }

	} //switch (function) {

	return false;
}

// ----------------------------------------------------------------------------
// This routine is called when a joystick is moved.
void Game_client::handle_joyaxis(int joystick, int axis, int value)
{
  // For now, just treat it like a key depending on <,=,>0. (with some
  // tolerance)

  value/=10; // Hopefully should drown out jitter for analogs.

  for(int player=0; player<nplayers; ++player)
    {
      if(players[player].joystick==joystick)
	{
	  Controls &controls=players[player].controls;

	  // X axis: axis 0
	  if(axis==0)
	    {
	      if(value==0)
		controls.clear_flag(Controls::cf_walk);
	      else
		controls.set_flag(Controls::cf_walk);

	      if(value>0)
		controls.set_flag(Controls::cf_direction_east);
	      else if(value<0)
		controls.clear_flag(Controls::cf_direction_east);
	    }
	  // Y axis
	  else if(axis==1)
	    {
	      // When do I need to do a controls_tick()?
	      if(value==0)
		{
		  players[player].elevate_up=false;
		  players[player].elevate_down=false;
		  if(!controls_ticked) controls_tick();
		}
	      else if(value>0)
		{
		  players[player].elevate_down=true;
		  players[player].elevate_up=false;
		  controls_ticked=false;
		}
	      else if(value<0)
		{
		  players[player].elevate_down=false;
		  players[player].elevate_up=true;
		  controls_ticked=false;
		}
	    }
	}
    }
}
