/***************************************************************************
                          game_client.h  -  description
                             -------------------
    begin                : Thu Feb 3 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#ifndef GAME_CLIENT_H
#define GAME_CLIENT_H

#include "configuration.h"
#include "clientworld.h"
#include "loader.h"
#include "controls.h"
#include "tcp_client.h"
#include "viewport.h"
#include "graphic_bar.h"
#include "graphic_msg.h"
#include "font_instance.h"
#include "keymapper.h"

// The maximum number of *local* players.
#define MAX_PLAYERS 4

class Game_client : public Configuration_observer {
public: 
	Game_client();
	~Game_client();

	//call to get things going:
	int init(Loader *loader_, Soundsystem *soundsystem_,
		 Configuration *configuration_);

	int setgeom(int xpos_, int ypos_, int view_xsize, int view_ysize, int view_quality_);

	// Must be called before addplayer.
	int join_game(const char *server, int server_port, int nplayers_,
		      const char *name, int color);

	// Adds a local player to the game.  FIXME: setup controls here.
	int addplayer(const char *player_name, int player_color);

	//void get_join_status(char *text, int textsize, int &progress_, int &max_progress_); remember to have a semaphor

	//poll once for each frame:
	//it gets new data from the server and
	int think();

	bool handle_input(Nil_inputevent *inputevent, int player);
	void handle_joyaxis(int joystick, int axis, int value);

	//call to have this game client draw it's view of the world
	int draw(Mutable_raw_surface *target);

	float get_trend_tick_duration() {return trend_tick_duration;}

	struct player_info
	{
	  int xpos; // These are offsets from our position.
	  int ypos;
	  int width;
	  int height;
	  Playername player_name;
	  int player_color;
	  int owner;
	  int joystick;
	  Controls controls;
	  bool elevate_up;
	  bool elevate_down;

		
	  Graphic_bar health_bar;
	  Graphic_bar weapon_bar[WEAPON_SLOTS];
		Graphic_msg frag_info;
		Graphic_msg message [ 3 ];

	  Viewport *viewport;
	};


	Clientworld clientworld;
	bool dropshadow;
	bool smooth;
	int last_update_size;
	int get_player_id ( int num ) {return players[ num ].owner;}
	player_info *get_player () {return &players[0];}

	void config_changed();

	// font for stats
	Font_instance *smallfont;
protected:

	// FIXME: use arbitrary number (STL vector?)
	int nplayers;
	player_info players[ MAX_PLAYERS ];

	float last_tick_time;
	float last_tick_duration;
	float trend_tick_duration;

	bool controls_ticked;
	void controls_tick();

	int xloc;
	int yloc;
	int width;
	int height;

	int map_xsize, map_ysize;

	int view_quality;
	bool bars_at_top;
	//void set_join_status(char *text, int &progress_, int &max_progress_);

	Tcp_client tcp_client;
	Loader *loader;
	Soundsystem *soundsystem;

private:

	void setup_viewports();
};

#endif

