/***************************************************************************
                          netpackets.h  -  description
                             -------------------
    begin                : Sun Jan 16 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
	structs used for client/server communication
 ***************************************************************************/

#ifndef _NETPACKETS_H_
#define _NETPACKETS_H_

#include "serializable.h"

#define PROTOCOL_SIGNATURE "NiL - The game"
#define PROTOCOL_VERSION "000215+multiplex"

typedef char Playername[40];

//-----------------------------------------------------------------------------
class Net_server_signature : public Serializable {
public:
	char signature[20];
	char version[20];
	
	Net_server_signature() {strcpy(signature,PROTOCOL_SIGNATURE); strcpy(version,PROTOCOL_VERSION);}
	bool server_ok() 	{return strcmp(signature,PROTOCOL_SIGNATURE) == 0;}
	bool version_ok() {return strcmp(version,PROTOCOL_VERSION) == 0;}

	virtual void serialize(Serializer *serializer) {
		serializer->rwstr(signature,sizeof(signature));
		serializer->rwstr(version,sizeof(version));
	}
};

//-----------------------------------------------------------------------------
//class Net_request
//class Net_request_info
//class Net_reply_info
//class Net_request_quit

// Hardcodes 4 as the maximum players/client!
//
//  The player info could be sent as a separate packet, which would provide
// more flexibility (players could, for instance, join and quit dynamically),
// but there seem to be race conditions that make that complicated.  Possible
// future extension.
class Net_joinrequest : public Serializable {
 public:
  struct player_info {
    int color;
    Playername name;
  };

  int nplayers;
  player_info players[4];

  virtual void serialize(Serializer *serializer)
  {
    serializer->rw(nplayers);
    for(int i=0; i<nplayers; ++i)
      {
	serializer->rw(players[i].color);
	serializer->rwstr(players[i].name, sizeof(players[i].name));
      }
  }
};

class Net_joinreply : public Serializable {
 public:
  int map_xsize;
  int map_ysize;
  unsigned int randomseed;

  bool zcompressed;
  unsigned int compressed_size; 	//number of bytes in the compressed image

  int player_ids[4];

  virtual void serialize(Serializer *serializer) {
    serializer->rw(map_xsize);
    serializer->rw(map_ysize);
    serializer->rw(randomseed);
    serializer->rw(zcompressed);
    serializer->rw(compressed_size);
    for(int i=0; i<4; ++i)
      serializer->rw(player_ids[i]);
  }
};

/*
	This is used to tell the other end of socket what kind of data comes next
*/
namespace NETSTATE {
	//The next lump of data will be...
  enum {
    CONTROLS,				//... a serialized Controls object
    WORLDUPDATE,		//... the combined state of a world
    UPDATEID,				//... a raw int telling us what updateid is interesting in the context it was sent

    DONE,         	//... Nothing, this set of lumps is done
    QUIT,						//... Nothing, terminate the connection _now_
    NETERROR					//... *Bugger*
  };
};

#endif

