/***************************************************************************
                          nil_server.cpp  -  description
                             -------------------
    begin                : Sun Nov 14 1999
    copyright            : (C) 1999 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

#include "systemheaders.h"
#include "nil_server.h"
#include "fluff.h"
#include "serverworld.h"
#include "loader.h"
#include "tcp_server.h"
#include "soundsystem.h"
#include "signal.h"

void run_server(Configuration &config)
{
	Weapon_style weapon_style;
	int gameweapons[WEAPON_SLOTS];
	setmsgsource("S");
	logmsg(lt_message,"Running server");

	Tcp_server tcp_server;
	
	signal( SIGPIPE, SIG_IGN );

	tcp_server.set_fudge(
		config.get_option_int("net_normal_backlog"),
		config.get_option_int("net_abnormal_backlog"),
		config.get_option_int("net_max_wait_time")
	);
	
	tcp_server.bind_and_listen(config.get_option_int("server_port"),config.get_option_int("net_multithread"));

	Loader loader;
	loader.addsources(config.get_option("data"));

	Soundsystem soundsystem(&loader, NULL);

	// Parse the weapon options (partly)
	if(!strcasecmp(config.get_option("weapon_style"), "randidentical"))
	  {
	    logmsg(lt_debug, "Using identically random weapon slots");
	    weapon_style=ws_randidentical;
	    for(int i=0; i<WEAPON_SLOTS; ++i)
	      gameweapons[i]=rnd(ot_weapon_first,ot_weapon_last);
	  }
	else if(!strcasecmp(config.get_option("weapon_style"), "rand"))
	  {
	    weapon_style=ws_rand;
	    logmsg(lt_debug, "Using random weapon slots");
	  }
	else
	  // Fixed or incorrect.
	  {
	    weapon_style=ws_fixed;
	    if(strcasecmp(config.get_option("weapon_style"), "fixed"))
	      logmsg(lt_warning, "Unknown weapon style %s, reverting to fixed", config.get_option("weapon_style"));

	    logmsg(lt_debug, "Using fixed weapon slots");

	    for(int i=0; i<WEAPON_SLOTS; ++i)
	      {
		char buf[64]; // way bigger than we need.

		snprintf(buf, 64, "weapon%i", i+1);
		buf[63]='\0'; // better safe than sorry.
		gameweapons[i]=ot_from_name(config.get_option(buf));

		if(gameweapons[i]==ot_error ||
		   gameweapons[i]<ot_weapon_first ||
		   gameweapons[i]>ot_weapon_last)
		  {
		    logmsg(lt_warning,
			   "Bad weapon name %s for slot %i, reverting to %s",
			   config.get_option(buf), i, OBJECT_NAMES[default_weapons[i]]);

		    gameweapons[i]=default_weapons[i];
		  }
	      }
	  }

	//set up the server:
	Serverworld serverworld(&loader,&soundsystem,
				config.get_option_int("always_reload"),
				weapon_style, gameweapons);
	tcp_server.serverworld = &serverworld;

	Config_xy world_size = config.get_option_xy("world_size");
	char *map_name = strdup ( config.get_option ( "map" ) );
	if ( strcmp ( map_name , "-=random=-" ) == 0 ) {
		serverworld.createmap(world_size.x,world_size.y,config.get_option_int("randomseed"));
	}
	else {
		serverworld.createmap ( map_name );
	}
	serverworld.save_original_map();
	logmsg(lt_debug,"map created");
	
	int traffic = 0;
	int frames = 0;
	bool done = false;

	unsigned int cur_time = time_ms();
	float gametime = 0.0;
	float stat_start = 0.0;
	unsigned int last_time = cur_time;
	int recycle_counter = 0;

  while (!done) {

		if (request_quit)
			done = true;

		{ //find the new time
			last_time = cur_time;
			cur_time = time_ms();
			unsigned int delta_time = time_elapsed(last_time,cur_time);
		
			gametime += delta_time/1000.0;
		}

		//see if it is time to recycle serializers
		if (recycle_counter-- <= 0) {
			tcp_server.recycle_serializers();
			//do a recycle run every 5 frames (maybe this shold be every frame, experiment)
			recycle_counter = 5;
		}

		serverworld.settime(gametime);
		serverworld.server_think(&tcp_server.update_queue);

		//update clients, update their controls and let new clients join.
		tcp_server.poll();

		//do some stats:
		traffic += serverworld.last_reader_length;
		frames++;
    //print stats:
		if (frames >= 150) {
			logmsg(lt_message,"Bytes/sec: %.1f, fps: %.2f, dt: %.2f, time:%.2f",
				traffic/(gametime-stat_start),
				frames/(gametime-stat_start),
				gametime-stat_start,
				gametime
			);
    	traffic = 0;
			frames = 0;
			stat_start = gametime;
		}

		//delay one tick
		while (time_elapsed(cur_time,time_ms()) < 40)
			usleep(1000);

  }

 	return;
}
