/***************************************************************************
                          sound.cpp  -  description
                             -------------------
    begin                : Thu Aug 15 2002
    copyright            : (C) 2002 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/

#include "systemheaders.h"
#include "sexchanger.h"
#include "fluff.h"
#include "sound.h"
#include "loader.h"
#include "nil_math.h"
#include "SDL_mixer.h"

// Currently only needed for MSB machines
#if __BYTE_ORDER == __BIG_ENDIAN
#include <endian.h>
#endif

/* Parts commented out because SDL_mixer doesn't support MSB wave files yet.
 * But the extra endian code should make it work on MSB machines. */

//-----------------------------------------------------------------------------
Sound::Sound()
{
	chunk = NULL;
}

//-----------------------------------------------------------------------------
Sound::~Sound()
{
	if (chunk) Mix_FreeChunk(chunk);
}

//-----------------------------------------------------------------------------
struct RIFF_Chunk {
	char id[4];
	Uint32 size;
};
#define idcmp(id,c1,c2,c3,c4) ((id)[0]==(c1) && (id)[1]==(c2) && (id)[2]==(c3) && (id)[3]==(c4))

struct WAVE_fmt {
	Uint16 encoding;	// encoding. 1=PCM/uncompressed
	Uint16 channels;	// output channels
	Uint32 samplerate;	// a.k.a. frequency, samples/sec
	Uint32 byterate;	// bytes/sec
	Uint16 blockalign;	// bytes/sample*channels
	Uint16 samplebits;	// bits/sample
	// Extra data follows in compressed/non-PCM formats
	// We don't care though.
};

//-----------------------------------------------------------------------------
bool Sound::load(Loader *loader, char *name, bool loadchunk)
{
	char *file = (char *)loader->getfile(name);
	if (!file) {
		logmsg(lt_error,"Unable to load sound %s!",name);
		return false;
	}

	int size = loader->filesize(name);
	if (size < 44) {
		logmsg(lt_error,"%s is too small to be a wave file.",name);
		return false;
	}

	char *f = file;
	struct RIFF_Chunk header;
//	bool msb; // big-endian?
	memcpy(&header,f,sizeof(struct RIFF_Chunk));
//	if (idcmp(header.id,'R','I','F','F')) // little-endian
//		msb = 0;
//	else if (idcmp(header.id,'R','I','F','X')) // big-endian
//		msb = 1;
//	else {
	if (!idcmp(header.id,'R','I','F','F')) {
		logmsg(lt_error,"%s has no RIFF header, not a wave file.",name);
		return false;
	}

#if __BYTE_ORDER == __BIG_ENDIAN
//	if (!msb) header.size = bytesex_change(header.size);
	header.size = bytesex_change(header.size);
//#else
//	if (msb) header.size = bytesex_change(header.size);
#endif

	if (header.size+8 != size)
		logmsg(lt_warning,"File size and size in %s's header do not match.",name);

	f += 8;
	if (!idcmp((char *)f,'W','A','V','E')) {
		logmsg(lt_error,"%s is not a wave file.",name);
		return false;
	}

	int offset = 12;
	f += 4;
	int datasize = 0;
	struct WAVE_fmt format;
	char found;
	while (offset < size) {
		memcpy(&header,f,sizeof(struct RIFF_Chunk));

#if __BYTE_ORDER == __BIG_ENDIAN
//		if (!msb) header.size = bytesex_change(header.size);
		header.size = bytesex_change(header.size);
//#else
//		if (msb) header.size = bytesex_change(header.size);
#endif

		if (idcmp(header.id,'f','m','t',' ')) {
			found |= 0x01;
			memcpy(&format,f+8,sizeof(struct WAVE_fmt));
		} else if (idcmp(header.id,'d','a','t','a')) {
			found |= 0x02;
			datasize += header.size;
		}

		offset += 8+header.size;
		f = file+offset;
	}

	if (!(found & 0x01)) {
		logmsg(lt_error,"Wave file %s is corrupt: no format chunk.",name);
		return false;
	}
	if (!(found & 0x02))  {
		logmsg(lt_error,"Wave file %s is corrupt: no data chunk.",name);
		return false;
	}

#if __BYTE_ORDER == __BIG_ENDIAN
//	if (!msb) format.byterate = bytesex_change(format.byterate);
	format.byterate = bytesex_change(format.byterate);
//#else
//	if (msb) format.byterate = bytesex_change(format.byterate);
#endif

	length = (float)datasize/(float)format.byterate;

	if (!loadchunk) return true;

	chunk = Mix_LoadWAV_RW(SDL_RWFromMem(file,size),1);
	if (!chunk) {
		logmsg(lt_error,"Unable to load wave file %s: %s",name,Mix_GetError());
		return false;
	}

	return true;
}
