/***************************************************************************
                          soundsystem.h  -  description
                             -------------------
    begin                : Mon Feb 21 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#ifndef SOUNDSYSTEM_H
#define SOUNDSYSTEM_H

#include "serializer_writer.h"
#include "serializer_reader.h"
#include "sound_event.h"
#include "game_client.h"

/* Sound flags */
#define	SOUND_LOOPING	0x01
#define SOUND_GLOBAL	0x02
#define SOUND_STATIC	0x04
#define SOUND_OBJECT	0x08
#define SOUND_DONE	0x10
#define SOUND_DIRTY	0x20

class Loader;

class Soundsystem {
	public:

	Soundsystem(Loader *loader_, Game_client *game_client_);
	virtual ~Soundsystem();

	void server_think(Serializer_reader *serializer_reader);
	void client_think(Serializer_writer &serializer_writer);
	void cleanup();

	int start_sound(char *sample_name, int flags, float volume);
	int start_sound(char *sample_name, int flags, float volume, float x_pos, float y_pos);
	int start_sound(char *sample_name, int flags, float volume, int objectid);

	void stop_sound(int sound_handle);
	void sound_halted(int channel);

	int get_sample_id(char *sample_name);
	bool get_quietmode() {return quietmode;}
	int get_channels() {return channels;}
	Sound *getsound(char *name, bool loadchunk) {return loader->sounds.getsound(name,loadchunk);}

	protected:

	Game_client *game_client;
	Loader *loader;

	int channels;
	bool quietmode;	

	Sound_events_handle *sound_events_handle;
	Sound_handles_server *sound_handles_server;
	int next_sound_handle;

	int start_sound(Sound_event *sound_event);
	void update();
};

#endif

